import threading
import time
import logging

from .counters import counter_instances


counter_engines = []


def pusher_thread(commit_interval):
    while True:
        if counter_instances:
            logging.debug("Committing {} counters...".format(len(counter_instances)))

            for counter_instance in counter_instances:
                counter_instance.commit(counter_engines)

        time.sleep(commit_interval)


def init_counter_pusher(commit_interval=1):
    t = threading.Thread(target=pusher_thread, args=(commit_interval, ))
    t.daemon = True
    t.start()


def add_counter_engine(counter_engine):
    counter_engines.append(counter_engine)



import collections
import time
import threading


counter_instances = []


class Counter:
    def __init__(self, name, aggregator=None):
        self.name = name
        self.data = collections.defaultdict(list)
        self.aggregator = aggregator
        self.lock = threading.Lock()

        counter_instances.append(self)

    def get_tags_str(self, tags):
        return ",".join(["{}:{}".format(k, v) for k, v in sorted(tags.items())])

    def set(self, value, tags=None):
        raise NotImplemented()

    def commit(self, counter_engines):
        raise NotImplemented()


class SimpleCounter(Counter):
    def set(self, value, tags=None):
        # Acquire lock.
        self.lock.acquire()

        # Update data.
        tags = tags or {}
        key = self.get_tags_str(tags)
        self.data[key].append((time.time(), value, tags))

        # Release lock.
        self.lock.release()

    def commit(self, counter_engines):
        # Acquire lock.
        self.lock.acquire()
        try:
            # Generate metrics.
            metrics = []
            for tag_str, tag_data in self.data.items():
                if self.aggregator is None:
                    for ts, value, tags in tag_data:
                        metrics.append({
                            "metric": self.name,
                            "timestamp": int(ts * 1000),
                            "value": value,
                            "tags": tags,
                        })
                else:
                    agg_value = self.aggregator.aggregate([
                        value for _, value, tags in tag_data
                    ])

                    metrics.append({
                        "metric": self.name,
                        "timestamp": int(time.time() * 1000),
                        "value": agg_value,
                        "tags": tag_data[0][2],
                    })

            # Commit metrics.
            for counter_engine in counter_engines:
                counter_engine.commit_metrics(metrics)
        finally:
            # Clear data.
            self.data.clear()

            # Reset aggregator.
            self.aggregator.reset()

            # Release lock.
            self.lock.release()



from prometheus_client import CollectorRegistry, Counter, Histogram, push_to_gateway
import json
import logging
import requests
import time


class CounterEngine:
    def commit_counter(self, counter):
        raise NotImplemented()


class OpenTSDBCounterEngine(CounterEngine):
    def __init__(self, hostname, port=4242):
        self.hostname = hostname
        self.port = port
        self.session = requests.Session()

    def commit_metrics(self, metrics):
        if not metrics:
            return True

        logging.debug("commiting metrics {} to opentsdb".format(metrics))

        res = self.session.post(
            "http://{}:{}/api/put?details=true".format(self.hostname, self.port),
            data=json.dumps(metrics),
        )
        return res.status_code == 200


class PrometheusCounterEngine(CounterEngine):
    def __init__(self, hostname, port=80, job_name="tscounters"):
        self.hostname = hostname
        self.port = port
        self.job_name = job_name

        self.registry = CollectorRegistry()
        self.counters = {}

    def get_counter(self, counter_name, tags):
        key = "{}({})".format(counter_name, ",".join(tags))

        if key not in self.counters:
            logging.info("creating counter {} with tags {}".format(counter_name, tags))
            self.counters[key] = Counter(
                counter_name,
                counter_name,
                tags,
                registry=self.registry,
            )

        return self.counters[key]

    def commit_metrics(self, metrics):
        if not metrics:
            return True

        logging.info("commiting metrics {} to prometheus".format(metrics))

        for metric in metrics:
            name = metric["metric"]
            tags = metric["tags"]
            value = metric["value"]

            counter = self.get_counter(name, list(tags.keys()))
            counter.labels(**tags).inc(value)

        push_to_gateway(
            "http://{}:{}".format(self.hostname, self.port),
            job=self.job_name,
            registry=self.registry,
        )

        return True



import time


class Aggregator:
    def aggregate(self, values):
        raise NotImplemented()

    def reset(self):
        pass


class SumAggregator(Aggregator):
    def aggregate(self, values):
        return sum(values)


class AvgAggregator(Aggregator):
    def aggregate(self, values):
        if not values:
            return 0
        else:
            return sum(values) / len(values)


class RateAggregator(Aggregator):
    def __init__(self):
        self.window_start_ts = time.time()

    def aggregate(self, values):
        window_end_ts = time.time()
        return sum(values) / (window_end_ts - self.window_start_ts)

    def reset(self):
        self.window_start_ts = time.time()



from .aggregator import SumAggregator, AvgAggregator, RateAggregator
from .counter_engine import OpenTSDBCounterEngine
from .counter_pusher import init_counter_pusher, add_counter_engine
from .counters import SimpleCounter


def init(commit_interval=1):
    init_counter_pusher(commit_interval=commit_interval)



#!/usr/bin/env python3

from setuptools import setup

setup(
    name='tscounters-py-nb',
    version='0.0.4',
    description='Time-series counters',
    author='Songyan Hou',
    author_email='songyan.hou@newsbreak.com',
    packages=['tscounters'],
    url="https://github.com/ParticleMedia/tscounters-py",
    keywords="opentsdb, metrics, counters",
    install_requires=[
        "requests",
        "prometheus_client",
    ]
)
